#include <sys/stat.h>

#include <fstream>
#include <sstream>

#include "general/AtExit.hh"
#include "general/autoarray.hh"
#include "general/fstream.hh"
#include "general/unittest.h"

#include "framecpp/Common/MemoryBuffer.hh"
#include "framecpp/Common/FrameBuffer.hh"

#include "framecpp/IFrameStream.hh"
#include "framecpp/OFrameStream.hh"
#include "framecpp/FrameH.hh"
#include "framecpp/FrRawData.hh"


using namespace FrameCPP;

using FrameCPP::Common::FrameBuffer;
using FrameCPP::Common::MemoryBuffer;
using FrameCPP::FrameH;
using FrameCPP::FrRawData;
using FrameCPP::FrAdcData;

typedef FrameBuffer< General::filebuf > frame_buffer_type;

General::UnitTest	Test;

//-----------------------------------------------------------------------
// Common tasks to be performed when exiting.
//-----------------------------------------------------------------------
inline void
depart( int ExitCode )
{
  General::AtExit::Cleanup( );
  exit( ExitCode );
}

inline const std::string&
filename_source( int Spec = FRAME_SPEC_CURRENT )
{
  static std::string filename;

  std::ostringstream filename_oss;

  filename_oss << "Z-R_std_test_frame_ver" << Spec
	       << "-" << 600000000
	       << "-" << 1
	       << ".gwf"
    ;
  
  filename = filename_oss.str( );

  return filename;
}

void
test_memory_buffer( )
{
  for ( INT_4U file_source_version = FRAME_SPEC_MIN,
	  file_source_version_end = FRAME_SPEC_MAX;
	file_source_version <= file_source_version_end;
	++file_source_version )
  {
    if ( file_source_version == 5 )
    {
      continue;
    }

    std::ostringstream	header;

    header << "test_memory_buffer:"
	   << " file source version: " << file_source_version
      ;

    try
    {
      const std::string
	filename( filename_source( file_source_version ) );

      struct stat		stat_buf;
  
      stat( filename.c_str( ), &stat_buf );

      MemoryBuffer		mb( std::ios::in );
      General::AutoArray< char >
	buf( new char[ stat_buf.st_size ] );
      mb.pubsetbuf( buf.get( ), stat_buf.st_size );
      std::ifstream	s;
      s.open( filename.c_str( ) );
      s.read( buf.get( ), stat_buf.st_size );
      s.close( );

      IFrameStream		ifs( false, &mb );

      IFrameStream::frame_h_type	frame( ifs.ReadNextFrame( ) );
      FrameH::rawData_type		rd( frame->GetRawData( ) );
      if ( rd )
      {
	Test.Check(  rd->RefFirstAdc( ).size( ) > 0 )
	  << header.str( )
	  << ": firstAdc count: " <<  rd->RefFirstAdc( ).size( )
	  << std::endl
	  ;
      }
      Test.Check(  frame->RefProcData( ).size( ) > 0 )
	<< header.str( )
	<< ": FrProcData count: "
	<< frame->RefProcData( ).size( )
	<< std::endl
	;
    }
    catch ( const std::exception& E )
    {
      Test.Check( false )
	<< header.str( )
	<< ": Caught an exception: " << E.what( )
	<< std::endl
	;
    }
    catch ( ... )
    {
      Test.Check( false )
	<< header.str( )
	<< ": Caught an unknown exception"
	<< std::endl
	;
    }
  }
}

void
test_standard( )
{
  try
  {
    //-------------------------------------------------------------------
    // input buffer needed
    //-------------------------------------------------------------------
    std::auto_ptr< frame_buffer_type >
      i_frame_buf( new frame_buffer_type( std::ios::in ) );

    std::auto_ptr< frame_buffer_type >
      o_frame_buf( new frame_buffer_type( std::ios::out ) );

    //-------------------------------------------------------------------
    // Calculate the filename
    //-------------------------------------------------------------------
    std::ostringstream	filename;

    filename << "Z-R_std_test_frame_ver" << FRAME_SPEC_CURRENT
	     << "-" << 600000000
	     << "-" << 1
	     << ".gwf"
      ;
    std::cerr << "DEBUG: source filename: " << filename.str( )
	      << std::endl
      ;

    std::ostringstream	ofilename;

    ofilename << "Z-R_std_test_frame_mem_buf_ver" << FRAME_SPEC_CURRENT
	      << "-" << 600000000
	      << "-" << 1
	      << ".gwf"
      ;
    //-----------------------------------------------------------------
    // Open the buffer and the stream
    //-----------------------------------------------------------------
    MemoryBuffer	omb( std::ios::out );

    i_frame_buf->open( filename.str( ).c_str( ),
		       std::ios::in | std::ios::binary );

    o_frame_buf->open( ofilename.str( ).c_str( ),
		       std::ios::out | std::ios::binary );

    IFrameStream	ifs( i_frame_buf.release( ) );

    OFrameStream	ofs_mem( false, &omb );

    try
    {
      while( 1 )
      {
	IFrameStream::frame_h_type	f = ifs.ReadNextFrame( );
		   
	ofs_mem.WriteFrame( f );
      }
    }
    catch( const std::out_of_range& E )
    {
      ofs_mem.Close( );

      OFrameStream	ofs( o_frame_buf.release( ) );

      //-----------------------------------------------------------------
      // load the input buffer with the output from the previous read
      //-----------------------------------------------------------------
      MemoryBuffer	imb( std::ios::in );
      imb.str( omb.str( ) );

      IFrameStream	ifs_mem( false, &imb );

      try
      {
	while( 1 )
	{
	  IFrameStream::frame_h_type	f = ifs_mem.ReadNextFrame( );
	  ofs.WriteFrame( f );
	}
      }
      catch( const std::out_of_range& E )
      {
	ofs.Close( );
      }
    
    }
    catch( std::exception& Exception )
    {
      std::cerr << "DEBUG:"
		<< " IStream: " << filename.str( )
		<< " : " << Exception.what( )
		<< std::endl
	;
      throw;
    }

  }
  catch ( const std::exception& E )
  {
    Test.Check( false ) << "Caught an exception: " << E.what( ) << std::endl;
  }
  catch ( ... )
  {
    Test.Check( false ) << "Caught an unknown exception" << std::endl;
  }
}

int
main( int ArgC, char**ArgV )
{
  //---------------------------------------------------------------------
  // Initialize the testing framework
  //---------------------------------------------------------------------
  Test.Init( ArgC, ArgV );

  //---------------------------------------------------------------------
  //
  //---------------------------------------------------------------------
  // test_standard( );
  test_memory_buffer( );

  //---------------------------------------------------------------------
  // Testing is complete. Report back to the user
  //---------------------------------------------------------------------
  try
  {
    //-------------------------------------------------------------------
    // Call cleanup code to ensure all of that is also working
    //-------------------------------------------------------------------
    General::AtExit::Cleanup( );
  }
  catch( ... )
  {
  }

  Test.Exit();
}
